#lang typed/racket

;; this module provides functions that parse ebml byte-strings

(provide ebml-read ebml-read/optimistic Element)

(define-type Element (Rec EE (List Natural (U Bytes (Listof EE)))))

;; a "front door" function that handles paths as
;; well as byte-strings
(: ebml-read : (U Bytes Path-String Input-Port) -> (Listof Element))
(define (ebml-read in-thing)
  (cond [(bytes? in-thing) (parse-elements 
                            (open-input-bytes in-thing))]
        [(path-string? in-thing)
         (define in-port (open-input-file in-thing))
         (begin0 (parse-elements in-port)
                 (close-input-port in-port))]
        [(port? in-thing)
         (parse-elements in-thing)]
        [else
         (raise-argument-error 'ebml-read
                               "byte-string or file path or port"
                               0 in-thing)]))

;; a "front door" function that handles paths as
;; well as byte-strings. The "optimistic" version assumes
;; that anything that can be parsed as ebml should be, which will
;; be incorrect for certain representations.

(: ebml-read/optimistic : (U Bytes Path-String Input-Port) -> (U Bytes (Listof Element)))
(define (ebml-read/optimistic in-thing)
  (cond [(bytes? in-thing) (optimistic-parse 
                            (open-input-bytes in-thing)
                            #f)]
        [(path-string? in-thing)
         (define in-port (open-input-file in-thing))
         (begin0 (optimistic-parse in-port #f)
                 (close-input-port in-port))]
        [(port? in-thing)
         (optimistic-parse in-thing #f)]
        [else
         (raise-argument-error 'ebml-read/optimistic
                               "byte-string or file path"
                               0 in-thing)]))



;; parse elements until running out of bytes; signal an error if
;; the byte-string doesn't end at the end of an element.
(: parse-elements : Input-Port -> (Listof Element))
(define (parse-elements port)
  (cond [(eof-object? (peek-byte port)) empty]
        [else
         (cons (parse-element port)
               (parse-elements port))]))

;; the optimistic parser just assumes that every element
;; contains a list of other elements, and falls back
;; to a byte string when parsing it signals an error.
;; this is an error-prone way to parse, because you may
;; have, e.g., a string that just happens to look like
;; a piece of encoded data. For better control, don't
;; use the optimistic-parser.

(: optimistic-parse : Input-Port (U #f Bytes) -> (U Bytes (Listof Element)))
(define (optimistic-parse port fallback-bytes)
  (with-handlers ([exn:fail? (lambda ([exn : exn])
                               (or fallback-bytes
                                   (raise exn)))])
    (define sub-elements (ebml-read port))
    (for/list: : (Listof Element) ([elt : Element sub-elements])
      (match-define (list header-id (? bytes? body-bytes)) elt)
      (list header-id (optimistic-parse 
                       (open-input-bytes body-bytes)
                       body-bytes)))))

;; bytes? fix? -> (list/c ebml-element? fix?)
;; given a byte string and a location, parse a single ebml element and 
;; return the element and the new parse position
(: parse-element : Input-Port -> Element)
(define (parse-element port)
  (define first-byte (assert (read-byte port) byte?))
  (match-define (list header-id-first-byte header-len) 
    (cond [(not (= 0 (bitwise-and #x80 first-byte)))
           (list (bitwise-and #x7F first-byte) 1)]
          [(not (= 0 (bitwise-and #x40 first-byte)))
           (list (bitwise-and #x3F first-byte) 2)]
          [(not (= 0 (bitwise-and #x20 first-byte)))
           (list (bitwise-and #x1F first-byte) 3)]
          [(not (= 0 (bitwise-and #x10 first-byte)))
           (list (bitwise-and #x0F first-byte) 4)]
          [else (error 'parse-element 
                       "not a legal header-id beginning byte: ~s"
                       first-byte)]))
  (define header-followup-bytes 
    (assert (read-bytes (sub1 header-len) port) bytes?))
  (when (< (bytes-length header-followup-bytes) (sub1 header-len))
    (raise-argument-error 'parse-element 
                          "port containing complete header id"
                          0 port))
  (define other-bits (bytes->list header-followup-bytes))
  (define header-id (+ (arithmetic-shift header-id-first-byte
                                         (* (length other-bits) 8))
                        (bytes->uint other-bits 0)))
  (cond [(= header-id (sub1 (expt 2 (* header-len 7))))
         (raise-result-error 'parse-element
                             "non-reserved header id"
                             header-id)])
  (parse-data-len port header-id))


;; port? fix? -> ebml-element?
;; given a byte string, a location, and a header ID, parse the data length
;; and data and return a Parse-Result
(: parse-data-len : Input-Port Natural -> Element)
(define (parse-data-len port header-id)
  (define first-byte (assert (read-byte port) byte?))
  (match-define (list data-len-first-byte data-len-len)
    (cond [(not (= 0 (bitwise-and #x80 first-byte)))
           (list (bitwise-and #x7F first-byte) 1)]
          [(not (= 0 (bitwise-and #x40 first-byte)))
           (list (bitwise-and #x3F first-byte) 2)]
          [(not (= 0 (bitwise-and #x20 first-byte)))
           (list (bitwise-and #x1F first-byte) 3)]
          [(not (= 0 (bitwise-and #x10 first-byte)))
           (list (bitwise-and #x0F first-byte) 4)]
          [(not (= 0 (bitwise-and #x08 first-byte)))
           (list (bitwise-and #x07 first-byte) 5)]
          [(not (= 0 (bitwise-and #x04 first-byte)))
           (list (bitwise-and #x03 first-byte) 6)]
          [(not (= 0 (bitwise-and #x02 first-byte)))
           (list (bitwise-and #x01 first-byte) 7)]
          [(not (= 0 (bitwise-and #x01 first-byte)))
           (list (bitwise-and #x00 first-byte) 8)]
          [else (error 'parse-data-len
                       "not a legal data-len beginning byte: ~s" first-byte)]))
  (define followup-data-len-bytes 
    (assert (read-bytes (sub1 data-len-len) port) bytes?))
  (when (< (bytes-length followup-data-len-bytes) 
           (sub1 data-len-len))
    (raise-argument-error
     'parse-data-len "port containing complete data length"
     0 port header-id))
  (define other-bits (bytes->list followup-data-len-bytes))
  (define data-len (+ (arithmetic-shift data-len-first-byte
                                         (* (length other-bits) 8))
                       (bytes->uint other-bits 0)))
  (cond [(= data-len (sub1 (expt 2 (* data-len-len 7))))
         (error "reserved data length id: ~s for data length len: ~s" data-len data-len-len)])
  (define data-bytes (assert (read-bytes data-len port) bytes?))
  (when (< (bytes-length data-bytes) 
           (sub1 data-len))
    (raise-argument-error
     'parse-data-len "port containing complete data"
     0 port header-id))
  (list header-id data-bytes))



;; parses a byte string as an unsigned integer. Works for lists
;; of funny lengths such as three or five.
;; (listof byte?) fix? -> fix?
(: bytes->uint : (Listof Byte) Natural -> Natural)
(define (bytes->uint l accum)
  (cond [(empty? l) accum]
        [else (bytes->uint (rest l) 
                           (+ (first l) (arithmetic-shift accum 8)))]))


;; uncomment after bug is fixed ...
#;(module+ test
  (require typed/rackunit)
  (check-equal? 
   (parse-element 
    (open-input-bytes (bytes #b10110000 #b10000001 13)))
   (list #x30 (bytes 13))))


#;(module+ test
  (check-exn 
   (lambda (exn) #t)
   (lambda () (parse-element (open-input-bytes (bytes #xFF)))))
  (check-equal?
   (parse-element (open-input-bytes
                   (bytes #b10000110 #b10000011 34 27 97)))
   (list 6 (bytes 34 27 97)))
  (check-equal? 
   (parse-element 
    (open-input-bytes
     (bytes #b01000000 #b00000110 #b10000011 34 27 97)))
   (list 6 (bytes 34 27 97)))
  
  (check-equal? 
   (parse-elements
    (open-input-bytes
     (bytes #b01000000 #b00000110 #b10000011 34 27 97)))
   (list (list 6 (bytes 34 27 97))))
  
  (check-equal? 
   (ebml-read 
    (open-input-bytes
     (bytes #b01000000 #b00000110 #b10000011 34 27 97)))
   (list (list 6 (bytes 34 27 97))))
  
  (check-equal? 
   (ebml-read (bytes #b01000000 #b00000110 #b10000011 34 27 97
                     #b10000011 #b10001011 1 2 3 4 5 6 7 8 9 10 11))
   (list (list 6 (bytes 34 27 97))
         (list 3 (bytes 1 2 3 4 5 6 7 8 9 10 11))))
  
  (check-equal? 
   (ebml-read (bytes #b01000000 #b00000110 #b10000011 34 27 97
                     #b10000011 #b10001011 
                     #b10000010 #b10000011 1 2 3
                     #b10000100 #b10000100 1 2 3 4))
   (list (list 6 (bytes 34 27 97))
         (list 3 (bytes #b10000010 #b10000011 1 2 3
                        #b10000100 #b10000100 1 2 3 4))))
  
  (check-equal? 
   (ebml-read/optimistic
    (bytes #b01000000 #b00000110 #b10000011 34 27 97
           #b10000011 #b10001011 
           #b10000010 #b10000011 1 2 3
           #b10000100 #b10000100 1 2 3 4))
   (list (list 6 (bytes 34 27 97))
         (list 3 (list (list 2 (bytes 1 2 3))
                       (list 4 (bytes 1 2 3 4)))))))




